/*
 * Author: Tony Olech - Elan Digital Systems Limited
 * $Revision: 1.11 $
 * $Date: 2007/11/15 16:07:38 $
 * 
 * PCMCIA driver for VMB USB HUB
 * Filename: vmb_cs.c
 * Email:    tony.olech@elandigitalsystems.com
 *
 *  16-bit pcmcia driver for the
 *  VMB 2 channel USB Host Controller
 *
 *  Interfaces to kernel 2.6.19
 *
 *  Structure based on the pcmcia driver for SL811
 *
 *  The VMB is not a card-bus device
 *
 * Started on: 2005-09-19
 */

#include <linux/version.h>
#include <linux/kernel.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#error you need at least 2.6.13 to compile the driver
#endif
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>
#include <pcmcia/ss.h>

#include "/usr/src/linux/drivers/pcmcia/cs_internal.h"

#include "usb_vmb.h"

#define DRIVER_VERSION "$Revision: 1.11 $"

MODULE_AUTHOR("Tony Olech");
MODULE_VERSION(DRIVER_VERSION);
MODULE_DESCRIPTION("PCMCIA VMB USB Host Controller driver for 2.6.19+");
MODULE_LICENSE("GPL");

/*====================================================================*/
/* MACROS                                                             */
/*====================================================================*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
#define CONFIG_RESOURCES_64BIT
#else
#undef CONFIG_RESOURCES_64BIT
#endif

#if defined(DEBUG) || defined(CONFIG_USB_DEBUG) || defined(PCMCIA_DEBUG)

#define LOCK_DEBUGGING_COMPILED_IN 0
#if defined(DEBUG) || defined(CONFIG_USB_DEBUG) || defined(PCMCIA_DEBUG)

#define DEBUGGING_COMPILED_IN 1

#else
#define DEBUGGING_COMPILED_IN 0
#endif

static int pc_debug = 1;
module_param(pc_debug, int, 0644);

#define DBG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG "vmb_cs: " args)
static int max_assert_reports = 100;
#define ASSERT(cond)    if (unlikely(!(cond)) && max_assert_reports > 0) {max_assert_reports--;printk(KERN_ERR "ASSERTION" __stringify(cond) " failure in %s at %s:%d\n",__FUNCTION__,__FILE__,__LINE__);dump_stack();}

#else
#define DBG(n, args...) do{}while(0)
#endif	/* no debugging */

#define ERR(stuff...)     printk(KERN_ERR "vmb_cs: " stuff)
#define WARN(stuff...)    printk(KERN_WARNING "vmb_cs: " stuff)
#define INFO(stuff...)    printk(KERN_INFO "vmb_cs: " stuff)

#define INT_MODULE_PARM(n, v) static int n = v; module_param(n, int, 0444)

#define CS_CHECK(fn, ret) \
	do { \
		last_fn = (fn); \
		if ((last_ret = (ret)) != 0) \
			goto cs_failed; \
	} while (0)

/*====================================================================*/
/* VARIABLES                                                          */
/*====================================================================*/

extern struct platform_driver vmb_driver;

static const char driver_name[DEV_NAME_LEN]  = "vmb_cs";

typedef struct local_info_t {
	struct pcmcia_device *p_dev;	
	dev_node_t		node;
	struct vmb_platform_data platform_data;
	struct resource resources[3];
	struct platform_device platform_dev;
} local_info_t;

/*====================================================================*/

static void release_platform_dev(struct device * dev)
{
	DBG(0, "%08lX %s()\n",jiffies,__FUNCTION__);
	dev->parent = NULL;
	DBG(0, "%08lX %s() completed\n",jiffies,__FUNCTION__);
}

/*
 * This is the last stage of a successful card probe
 *
 */
static int vmb_hcd_init(local_info_t*local_info, struct device *parent, int irq, struct resource*io_res, struct resource*mem_res)
{
#ifdef CONFIG_RESOURCES_64BIT
	DBG(0, "%s(%d,%lx-%lx,%lx-%lx)\n",__FUNCTION__,irq,io_res->start,io_res->end,mem_res->start,mem_res->end);
#else	
	DBG(0, "%s(%d,%x-%x,%x-%x)\n",__FUNCTION__,irq,io_res->start,io_res->end,mem_res->start,mem_res->end);
#endif
	if (local_info->platform_dev.dev.parent)
		return -EBUSY;

	local_info->platform_data.ports = 2;
	local_info->platform_data.potpg = 100;
	local_info->platform_data.power = 1000; /* == 1000mA */
	local_info->platform_data.reset = NULL; //... FIXME:  invoke CF reset on the card

	local_info->platform_dev.id            = -1;
	local_info->platform_dev.resource      = local_info->resources;
	local_info->platform_dev.num_resources = ARRAY_SIZE(local_info->resources);
	local_info->platform_dev.dev.platform_data = &local_info->platform_data;
	local_info->platform_dev.dev.parent        = parent;
	local_info->platform_dev.dev.release       = release_platform_dev;
	local_info->platform_dev.dev.dma_mask		= 0;	// we do not do DMA

	local_info->platform_dev.name = vmb_driver.driver.name;

	local_info->resources[0].name = "VMB-IRQ";
	local_info->resources[0].flags = IORESOURCE_IRQ;
	ASSERT(NULL == local_info->resources[0].parent);
	local_info->resources[0].start = irq;
	local_info->resources[0].end = irq;

	local_info->resources[1].name = "VMB-IO";
	local_info->resources[1].flags = IORESOURCE_IO;
	ASSERT(NULL == local_info->resources[1].parent);
	local_info->resources[1].start = io_res->start;
	local_info->resources[1].end = io_res->end;

	local_info->resources[2].name = "VMB-MEM";
	local_info->resources[2].flags = IORESOURCE_MEM;
	ASSERT(NULL == local_info->resources[2].parent);
	local_info->resources[2].start = mem_res->start;
	local_info->resources[2].end   = mem_res->end;

	if (local_info->p_dev->has_manf_id && 0x0355 == local_info->p_dev->manf_id
	&&  local_info->p_dev->has_card_id && 0x4c45 == local_info->p_dev->card_id
	   ) {
		local_info->platform_data.switch_on_required = 1;
	} else {	
		local_info->platform_data.switch_on_required = 0;
	}

	{
		int result = platform_device_register(&local_info->platform_dev);
		DBG(0, "%08lX %s() completed %d\n",jiffies,__FUNCTION__,result);
		return result;
	}
}
/*====================================================================*/

#if DEBUGGING_COMPILED_IN
extern void vmb_hcd_switch_on_diagnostics(void);
#endif
static void vmb_cs_release(struct pcmcia_device *p_dev)
{
	local_info_t *local_info = p_dev->priv;

	DBG(0, "%s(p_dev=0x%p)\n",__FUNCTION__, p_dev);

#if DEBUGGING_COMPILED_IN
	vmb_hcd_switch_on_diagnostics();
#endif

	pcmcia_disable_device(p_dev);

	DBG(0, "%s(local_info=0x%p platform_dev=0x%p)\n",__FUNCTION__,local_info,&local_info->platform_dev);

	platform_device_unregister(&local_info->platform_dev);

	DBG(0, "%s() completed\n",__FUNCTION__);
}

static void vmb_cs_remove(struct pcmcia_device *p_dev) // pcmcia_driver interface
{
	pc_debug = 0x7fffffff;

	DBG(0, "%s(p_dev=0x%p)\n",__FUNCTION__, p_dev);

	vmb_cs_release(p_dev);

	kfree(p_dev->priv);
	DBG(0, "%s() completed\n",__FUNCTION__);
}


static int vmb_cs_config(struct pcmcia_device *p_dev)
{
	local_info_t *local_info = p_dev->priv;
	struct device		*parent = &handle_to_dev(p_dev);
	tuple_t			tuple;
	cisparse_t		parse;
	int			last_fn, last_ret;
	win_req_t req;
	memreq_t map;
	u_char			buf[64];
	config_info_t		conf;
	cistpl_cftable_entry_t	dflt = { 0 };
	struct resource * io_res = NULL;
	struct resource * mem_res = NULL;

	DBG(0, "%08lX %s(0x%p)\n",jiffies,__FUNCTION__, p_dev);

	tuple.DesiredTuple = CISTPL_CONFIG;
	tuple.Attributes = 0;
	tuple.TupleData = buf;
	tuple.TupleDataMax = sizeof(buf);
	tuple.TupleOffset = 0;
	CS_CHECK(GetFirstTuple, pcmcia_get_first_tuple(p_dev, &tuple));
	CS_CHECK(GetTupleData, pcmcia_get_tuple_data(p_dev, &tuple));
	CS_CHECK(ParseTuple, pcmcia_parse_tuple(p_dev, &tuple, &parse));
	p_dev->conf.ConfigBase = parse.config.base;
	p_dev->conf.Present = parse.config.rmask[0];

	/* Look up the current Vcc */
	CS_CHECK(GetConfigurationInfo,
			pcmcia_get_configuration_info(p_dev, &conf));

	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	CS_CHECK(GetFirstTuple, pcmcia_get_first_tuple(p_dev, &tuple));
	while (1) {
		int err;
		cistpl_cftable_entry_t	*cfg = &(parse.cftable_entry);

		if (pcmcia_get_tuple_data(p_dev, &tuple) != 0
				|| pcmcia_parse_tuple(p_dev, &tuple, &parse)
						!= 0)
			goto next_entry;

		if (cfg->flags & CISTPL_CFTABLE_DEFAULT) {
			dflt = *cfg;
		}

		if (cfg->index == 0)
			goto next_entry;

		p_dev->conf.ConfigIndex = cfg->index;

		/* Use power settings for Vcc and Vpp if present */
		/*  Note that the CIS values need to be rescaled */
		if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM)) {
			if (cfg->vcc.param[CISTPL_POWER_VNOM]/10000
					!= conf.Vcc)
				goto next_entry;
		} else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM)) {
			if (dflt.vcc.param[CISTPL_POWER_VNOM]/10000
					!= conf.Vcc)
				goto next_entry;
		}

		if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
			conf.Vpp1 = conf.Vpp2 =
				cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
		else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
			conf.Vpp1 = conf.Vpp2 =
				dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;

		/* we need an interrupt */
		if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
			p_dev->conf.Attributes |= CONF_ENABLE_IRQ;

		/* IO window settings */
		p_dev->io.NumPorts1 = p_dev->io.NumPorts2 = 0;
		if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0)) {
			cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
			p_dev->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
			if (!(io->flags & CISTPL_IO_8BIT))
				p_dev->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
			if (!(io->flags & CISTPL_IO_16BIT))
				p_dev->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
			p_dev->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
			p_dev->io.BasePort1 = io->win[0].base;
			p_dev->io.NumPorts1 = io->win[0].len;
			if (io->nwin > 1) {
				p_dev->io.Attributes2 = p_dev->io.Attributes1;
				p_dev->io.BasePort2 = io->win[1].base;
				p_dev->io.NumPorts2 = io->win[1].len;
			}
			/* This reserves IO space but doesn't actually enable it */
			if ((err=pcmcia_request_io(p_dev, &p_dev->io)) != CS_SUCCESS)
			{
				if (io->nwin > 1)
					printk(KERN_DEBUG "pcmcia_request_io: base=%04X(%d),%04X(%d) aligned=%d FAILED %d\n"
						,p_dev->io.BasePort1,p_dev->io.NumPorts1,p_dev->io.BasePort2,p_dev->io.NumPorts2,p_dev->io.IOAddrLines,err);
				else
					printk(KERN_DEBUG "pcmcia_request_io: base=%04X(%d) aligned=%d FAILED %d\n"
						,p_dev->io.BasePort1,p_dev->io.NumPorts1,p_dev->io.IOAddrLines,err);
				goto next_entry;
			}
			else
			{
				io_res = p_dev->socket->io[0].res;

				if (io->nwin > 1)
				{
					printk(KERN_DEBUG "pcmcia_request_io: base=%04X(%d),%04X(%d) aligned=%d OK\n"
						,p_dev->io.BasePort1,p_dev->io.NumPorts1,p_dev->io.BasePort2,p_dev->io.NumPorts2,p_dev->io.IOAddrLines);
				}
				else
				{
					if (p_dev->io.NumPorts1)
							DBG(1,"IO %04x -- %04x\n",p_dev->io.BasePort1,p_dev->io.BasePort1+p_dev->io.NumPorts1-1);
					if (p_dev->io.NumPorts2)
							DBG(1,"IO %04x -- %04x\n",p_dev->io.BasePort2,p_dev->io.BasePort2+p_dev->io.NumPorts2-1);

					printk(KERN_DEBUG "pcmcia_request_io: base=%04X(%d) aligned=%d res=%p OK\n"
						,p_dev->io.BasePort1,p_dev->io.NumPorts1,p_dev->io.IOAddrLines,io_res);
				}
			}
		}
		/*
			Now set up a common memory window, if needed.  There is room
			in the pcmcia_device structure for one memory window handle,
			but if the base addresses need to be saved, or if multiple
			windows are needed, the info should go in the private data
			structure for this device.

			Note that the memory window base is a physical address, and
			needs to be mapped to virtual space with ioremap() before it
			is used.
		*/
		if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0)) {
			cistpl_mem_t *mem = (cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
			req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
			req.Attributes |= WIN_ENABLE;
			req.Base = mem->win[0].host_addr;
			req.Size = mem->win[0].len;
			if (req.Size < 0x1000) req.Size = 0x1000;
			req.AccessSpeed = 0;
			if ((err=pcmcia_request_window(&p_dev, &req, &p_dev->win)) != CS_SUCCESS)
			{
				printk(KERN_DEBUG "pcmcia_request_window: base=%08X(%u) FAILED %d\n",(u_int)req.Base,req.Size,err);
				goto next_entry;
			}
			else
			{
				mem_res = p_dev->win->ctl.res;
#ifdef CONFIG_RESOURCES_64BIT
				DBG(-1,"MEM %08lx --> %08lx  %08lx + %x\n",mem_res->start,mem_res->end,req.Base,req.Size);
#else				
				DBG(-1,"MEM %08x --> %08x  %08lx + %x\n",mem_res->start,mem_res->end,req.Base,req.Size);
#endif				
				printk(KERN_DEBUG "pcmcia_request_window: base=%08X(%u) res=%p OK\n",(u_int)req.Base,req.Size,mem_res);
			}
			map.Page = 0; map.CardOffset = mem->win[0].card_addr;
			if ((err=pcmcia_map_mem_page(p_dev->win, &map)) != CS_SUCCESS)
			{
				printk(KERN_DEBUG "pcmcia_map_mem_page: offset=%08X FAILED %d\n",map.CardOffset,err);
				goto next_entry;
			}
			else
			{
				printk(KERN_DEBUG "pcmcia_map_mem_page: offset=%08X OK\n",map.CardOffset);
			}
		}
		break;

next_entry:
		pcmcia_disable_device(p_dev);
		last_ret = pcmcia_get_next_tuple(p_dev, &tuple);
	}

	/* require an IRQ and seven registers */
	if (!p_dev->io.NumPorts1 || p_dev->io.NumPorts1 < 7)
		goto cs_failed;
	if (p_dev->conf.Attributes & CONF_ENABLE_IRQ)
		CS_CHECK(RequestIRQ, pcmcia_request_irq(p_dev, &p_dev->irq));
	else
		goto cs_failed;

	CS_CHECK(RequestConfiguration, pcmcia_request_configuration(p_dev, &p_dev->conf));

	sprintf(local_info->node.dev_name, driver_name);
	local_info->node.major = local_info->node.minor = 0;
	p_dev->dev_node = &local_info->node;

	printk(KERN_INFO "%s: index 0x%02x: Vcc %d.%d",
	       local_info->node.dev_name, p_dev->conf.ConfigIndex,
	       conf.Vcc/10, conf.Vcc%10);
	if (conf.Vpp1)
		printk(", Vpp %d.%d", conf.Vpp1/10, conf.Vpp1%10);
	printk(", irq %d", p_dev->irq.AssignedIRQ);
	printk(", io 0x%04x-0x%04x", p_dev->io.BasePort1, p_dev->io.BasePort1+p_dev->io.NumPorts1-1);
	if (p_dev->io.NumPorts2) printk(" & 0x%04x-0x%04x", p_dev->io.BasePort2, p_dev->io.BasePort2+p_dev->io.NumPorts2-1);
	if (p_dev->win) printk(", mem 0x%06lx-0x%06lx", req.Base, req.Base+req.Size-1);
	printk("\n");

	if (vmb_hcd_init(local_info, parent, p_dev->irq.AssignedIRQ, io_res, mem_res)
			< 0) {
		ERR("vmb_hcd_init failed\n");
cs_failed:
		ERR("%08lX %s() failed\n",jiffies,__FUNCTION__);
		cs_error(p_dev, last_fn, last_ret);
		vmb_cs_release(p_dev);
	}
	DBG(0, "%08lX %s() completed\n",jiffies,__FUNCTION__);
	return 0;
}

static int vmb_cs_probe(struct pcmcia_device *p_dev) // pcmcia_driver interface
{
	local_info_t *local;

	INFO("probing card with ID=%04x:%04x\n",p_dev->has_manf_id ? p_dev->manf_id : 0xFFFF,p_dev->has_card_id ? p_dev->card_id : 0xFFFF);

	local = kmalloc(sizeof(local_info_t), GFP_KERNEL);
	if (!local) {
		DBG(0, "%08lX %s() out of memory\n",jiffies,__FUNCTION__);
		return -ENOMEM;
	}
	memset(local, 0, sizeof(local_info_t));
	local->p_dev = p_dev;
	p_dev->priv = local;

	/* Initialize */
	p_dev->irq.Attributes = RES_SHARED; // was IRQ_TYPE_EXCLUSIVE;
	p_dev->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
	p_dev->irq.Handler = NULL;

	p_dev->conf.Attributes = 0;
	p_dev->conf.IntType = INT_MEMORY_AND_IO;

	return vmb_cs_config(p_dev);
}

static struct pcmcia_device_id vmb_ids[] = {
	PCMCIA_DEVICE_MANF_CARD(0x015d, 0x4c45), /* VMB Card */
	PCMCIA_DEVICE_MANF_CARD(0xc036, 0x1003), /* VMB inside NetIndex */
	PCMCIA_DEVICE_PROD_ID12("NetIndex","D01NX",0xf2ba4352,0x73bb32c2),
	PCMCIA_DEVICE_PROD_ID12("FOMA","M2501 USB Host Controller",0x81b58d52,0x6891e380),
	//PCMCIA_DEVICE_MANF_CARD(0x0355,0x4c45), /* VMB Inside FOMA*/
	PCMCIA_DEVICE_NULL,
};
MODULE_DEVICE_TABLE(pcmcia, vmb_ids);

static struct pcmcia_driver vmb_cs_driver = {
	.owner		= THIS_MODULE,
	.drv		= {
		.name	= (char *)driver_name,
	},
	.probe		= vmb_cs_probe,
	.remove		= vmb_cs_remove,
	.id_table	= vmb_ids,
};

/*====================================================================*/

static int __init init_vmb_cs(void)
{
	DBG(0, "%08lX %s()\n",jiffies,__FUNCTION__);
	INFO("driver %s %s pc_debug=%d built at %s on %s\n", driver_name, DRIVER_VERSION, pc_debug, __TIME__, __DATE__);
	{
		int result = pcmcia_register_driver(&vmb_cs_driver);
		DBG(0, "%08lX %s() completed %d\n",jiffies,__FUNCTION__,result);
		return result;
	}
}
module_init(init_vmb_cs);

static void __exit exit_vmb_cs(void)
{
	DBG(0, "%08lX %s()\n",jiffies,__FUNCTION__);
	pcmcia_unregister_driver(&vmb_cs_driver);
	DBG(0, "%08lX %s() completed\n",jiffies,__FUNCTION__);
}
module_exit(exit_vmb_cs);
